
from .model import ModelBase, make_declarative_base
from .query import Query
from .manager import Manager, ManagerBase

__version__ = '0.2.1'
__author__ = 'Derrick Gilland <dgilland@gmail.com>'
__all__ = ['ModelBase', 'make_declarative_base', 'Query', 'Manager', 'ManagerBase']

print locals()



from sqlalchemy import inspect, orm, and_, or_
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm.exc import UnmappedClassError

import query
from utils import classproperty, is_sequence

class ModelBase(object):
    '''Augmentable Base class for adding shared model properties/functions'''

    # default table args
    __table_args__ = {}

    # define a default order by when not specified by query operation
    # eg: { 'order_by': [column1, column2] }
    __mapper_args__ = {}

    # query class to use for self.query
    query_class = query.Query

    # an instance of `query_class`
    # can be used to query the database for instances of this model
    # @note: requires setting Class.query = QueryProperty(session) when session available
    # see `make_declarative_base()` for automatic implementation
    query = None

    # specify sqla.session.query.filter() options for advanced and simple searches
    # eg: { key: lambda value: Model.column_name == val }
    advanced_search_config = {}
    simple_search_config = {}

    def __init__(self, *args, **kargs):
        self.update(*args, **kargs)

    def __repr__(self):
        '''
        Default representation of model
        '''
        values = ', '.join(['{0}={1}'.format(c, repr(getattr(self, c))) for c in self.columns])
        return '<{0}({1})>'.format(self.__class__.__name__, values)

    def update(self, data_dict=None, strict=False, **kargs):
        '''
        Update model with arbitrary set of data
        '''

        data = data_dict if isinstance(data_dict, dict) else kargs

        updatable_fields = self.strict_update_fields if strict else data.keys()
        relationships = self.relationships

        for k,v in data.iteritems():
            if hasattr(self, k) and k in updatable_fields:
                # consider v a dict if any of its elements are a dict
                v_is_dict = any([isinstance(_v, dict) for _v in v]) if is_sequence(v) else isinstance(v, dict)

                attr = getattr(self, k)

                if hasattr(attr, 'update') and v_is_dict:
                    # nest calls to attr.update if available and input is a data dict
                    attr.update(v)
                else:
                    if k in relationships and v_is_dict and not v:
                        # typically, if v is {}, then we're usually updating a relationship attribute
                        # where the relationship has an empty/null value in the database
                        # (e.g. a frontend sends missing relationship attribute as {})
                        # but if we set a relationship attribute = {}, things blow up
                        # so instead, convert {} to None which is valid for standard relationship attribute
                        v = None
                    setattr(self, k, v)

    def to_dict(self):
        '''
        Return dict representation of model.

        Drill down to any relationships and serialize those too.

        Assume that the current object has been loaded (lazy/joined/etc)
        and don't try to expand anything, i.e., just serialize the currently loaded data
        '''
        d = {}

        for field, value in self.__dict__.iteritems():
            if field.startswith('_sa'):
                # skip sqlalchemy properties
                continue

            d[field] = [v.to_dict() for v in value] if is_sequence(value) else value

        return d

    @property
    def strict_update_fields(self):
        '''
        Model fields which are allowed to be updated during strict mode

        Default is to limit to table columns
        Override as needed in child classes
        '''
        return self.columns

    ##
    # search/filtering methods
    ##

    @classmethod
    def get_search(cls, search_dict, filter_fns):
        filters = []

        for key, filter_fn in [(k,v) for k,v in filter_fns.iteritems() if k in search_dict]:
            filters.append(filter_fn(search_dict[key]))

        return filters

    @classmethod
    def advanced_search(cls, search_dict):
        filters = None
        if cls.advanced_search_config:
            _filters = cls.get_search(search_dict, cls.advanced_search_config)

            if _filters:
                filters = and_(*_filters)

        return filters

    @classmethod
    def simple_search(cls, search_string):
        filters = None

        if cls.simple_search_config:
            terms = [s for s in search_string.split()]
            fields = cls.simple_search_config.keys()
            field_count = len(fields)

            search_filters = []

            for term in terms:
                # create a dict with each `config_search` key and `term` so filters can be applied to each combination
                # i.e. { config_search_key1: term, config_search_key2: term, ..., config_search_keyN, term }
                search_dict = dict(zip(fields, [term]*field_count))
                term_filters = cls.get_search(search_dict, cls.simple_search_config)

                if term_filters:
                    # `or` filters together since only 1 filter needs to match for `term`
                    search_filters.append(or_(*term_filters))

            if search_filters:
                # `and` all search conditions together
                # each `term` should have an OR'd set of filters that evaluates to True
                filters = and_(*search_filters)

        return filters

    ##
    # session based methods/properties
    ##

    @property
    def session(self):
        '''Return session belonging to self'''
        return orm.object_session(self)

    def flush(self, *args, **kargs):
        return self.session.flush([self], *args, **kargs)

    def delete(self, *args, **kargs):
        return self.session.delete(self, *args, **kargs)

    def expire(self, *args, **kargs):
        return self.session.expire(self, *args, **kargs)

    def refresh(self, *args, **kargs):
        return self.session.refresh(self, *args, **kargs)

    def expunge(self, *args, **kargs):
        return self.session.expunge(self, *args, **kargs)

    @classmethod
    def get(cls, *args, **kargs):
        return cls.query.get(*args, **kargs)

    ##
    # inspect based methods/properties
    ##

    @classproperty
    def attrs(cls):
        '''Return ORM attributes'''
        return inspect(cls).attrs.keys()

    @classproperty
    def descriptors(cls):
        '''Return all ORM descriptors'''
        return [d for d in inspect(cls).mapper.all_orm_descriptors.keys() if not d.startswith('__')]

    @classproperty
    def relationships(cls):
        '''Return ORM relationships'''
        return inspect(cls).mapper.relationships.keys()

    @classproperty
    def column_attrs(cls):
        '''Return table columns as list of class attributes at the class level'''
        return inspect(cls).mapper.column_attrs

    @classproperty
    def columns(cls):
        '''Return table columns'''
        return inspect(cls).mapper.columns.keys()

class QueryProperty(object):
    def __init__(self, session):
        self.session = session

    def __get__(self, model, Model):
        try:
            mapper = orm.class_mapper(Model)
            if mapper:
                if not getattr(Model, 'query_class', None):
                    Model.query_class = query.Query

                q = Model.query_class(mapper, session=self.session())
                q.__model__ = Model

                return q
        except UnmappedClassError:
            return None

def make_declarative_base(session=None, query_property=None, Model=None):
    Model = Model or declarative_base(cls=ModelBase, constructor=ModelBase.__init__)
    extend_declarative_base(Model, session, query_property)
    return Model

def extend_declarative_base(Model, session=None, query_property=None):
    # attach query attribute to Model if `session` object passed in
    if session:
        Model.query = query_property(session) if query_property else QueryProperty(session)





import re
from sqlalchemy.types import SchemaType, TypeDecorator, Enum

# declarative Enum type
class EnumSymbol(object):
    '''Define a fixed symbol tied to a parent class.'''

    def __init__(self, cls_, name, value, description):
        self.cls_ = cls_
        self.name = name
        self.value = value
        self.description = description

    def __reduce__(self):
        '''Allow unpickling to return the symbol linked to the DeclarativeEnum class.'''
        return getattr, (self.cls_, self.name)

    def __iter__(self):
        return iter([self.value, self.description])

    def __repr__(self):
        return "<%s>" % self.name

    def __str__(self):
        return self.name

class EnumMeta(type):
    '''Generate new DeclarativeEnum classes.'''

    def __init__(cls, classname, bases, dict_):
        cls._reg = reg = cls._reg.copy()
        for k, v in dict_.items():
            if isinstance(v, tuple):
                sym = reg[v[0]] = EnumSymbol(cls, k, *v)
                setattr(cls, k, sym)
        return type.__init__(cls, classname, bases, dict_)

    def __iter__(cls):
        return iter(cls._reg.values())

class DeclarativeEnumType(SchemaType, TypeDecorator):
    def __init__(self, enum):
        self.enum = enum
        self.impl = Enum(
            *enum.values(),
            name="ck%s" % re.sub('([A-Z])', lambda m:"_" + m.group(1).lower(), enum.__name__)
        )

    def _set_table(self, table, column):
        self.impl._set_table(table, column)

    def copy(self):
        return DeclarativeEnumType(self.enum)

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        return value.value

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return self.enum.from_string(value.strip())

class DeclarativeEnum(object):
    '''Declarative enumeration.'''

    __metaclass__ = EnumMeta
    _reg = {}

    @classmethod
    def from_string(cls, value):
        try:
            return cls._reg[value]
        except KeyError:
            raise ValueError("Invalid value for %r: %r" % (cls.__name__, value))

    @classmethod
    def values(cls):
        return cls._reg.keys()

    @classmethod
    def db_type(cls):
        return DeclarativeEnumType(cls)





from collections import Iterable

class classproperty(object):
    '''
    Decorator that adds class properties.
    Allows for usage like @property but applies the property at the class level.
    Helps avoid having to use metaclasses or other complex techniques to achieve similar results.
    '''
    def __init__(self, getter):
        self.getter = getter
    def __get__(self, instance, owner):
        return self.getter(owner)

def is_sequence(obj):
    '''
    Test if `obj` is an iterable but not dict or string
    '''
    return isinstance(obj, Iterable) and not isinstance(obj, basestring) and not isinstance(obj, dict)




from sqlalchemy import engine_from_config, orm
from sqlalchemy.orm.exc import UnmappedError

import query, model

class ManagerBase(object):
    '''Useful extensions to self.session'''

    def add(self, *instances):
        '''Override `session.add()` so it can function like `session.add_all()` and support chaining'''
        for instance in instances:
            if isinstance(instance, list):
                self.add(*instance)
            else:
                self.session.add(instance)
        return self.session

    def add_commit(self, *instances):
        '''Add instances to session and commit in one call'''
        self.add(*instances).commit()

    def delete(self, *instances):
        '''Override `session.delete()` so it can function like `session.add_all()` and support chaining'''
        for instance in instances:
            if isinstance(instance, list):
                self.delete(*instance)
            else:
                self.session.delete(instance)
        return self.session

    def delete_commit(self, *instances):
        '''Delete instances to session and commit in one call'''
        self.delete(*instances).commit()

class Manager(ManagerBase):
    '''
    Manager for session
    '''
    def __init__(self, Model=None, config=None, session=None, engine=None):
        self.session = None

        # declarative base class
        self.Model = Model or model.make_declarative_base()

        if config is None:
            config = {}

        self.init_session(config=config.get('session'), session=session)
        self.init_engine(config=config.get('engine'), engine=engine)

    def init_engine(self, config=None, engine=None, config_prefix=''):
        '''
        Initialize engine. Allow for lazy configuration after `__init__()`.
        If both `config` and `engine` are supplied, then the config generated engine will take precedence.
        New engine creation will also result in a new `session` object using `engine`.
        '''
        if config:
            engine = engine_from_config(config, prefix=config_prefix)

        self.init_session(config={'bind': engine})

    def init_session(self, config=None, session=None):
        '''
        Initialize session. Allow for lazy configuration after `__init__()`.
        If both `config` and `session` are supplied, `config` will configure `session`.
        '''
        self.session = session

        if self.session is None:
            self.session = orm.scoped_session(orm.sessionmaker())

        if config is None:
            config = {}

        config.setdefault('autocommit', False)
        config.setdefault('autoflush', True)
        config.setdefault('query_cls', query.Query)

        self.session.configure(**config)

        if self.Model:
            model.extend_declarative_base(self.Model, self.session)

    @property
    def metadata(self):
        return getattr(self.Model, 'metadata', None)

    @property
    def engine(self):
        return self.session.get_bind()

    def create_all(self):
        '''
        Creates database schema from models
        '''
        if self.metadata is None:
            raise UnmappedError('Missing declarative base model')
        self.metadata.create_all(bind=self.engine)

    def drop_all(self):
        '''
        Drops tables defined by models
        '''
        if self.metadata is None:
            raise UnmappedError('Missing declarative base model')
        self.metadata.drop_all(bind=self.engine)

    def __getattr__(self, attr):
        '''
        Delegate all other attributes to self.session
        '''
        return getattr(self.session, attr)




"""
alchy
-----

The SQLAlchemy enhancement library.

Documentation: https://github.com/dgilland/alchy
"""

from setuptools import setup

setup(
    name = 'alchy',
    version = '0.2.1',
    url = 'https://github.com/dgilland/alchy',
    license = 'MIT',
    author = 'Derrick Gilland',
    author_email = 'dgilland@gmail.com',
    description = 'SQLAlchemy enhancement library',
    long_description = __doc__,
    packages = ['alchy'],
    install_requires = [
        'SQLAlchemy>=0.9.0'
    ],
    test_suite = 'tests',
    keywords = 'sqlalchemy databases',
    classifiers = [
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
        'Topic :: Database :: Front-Ends'
    ]
)




from math import ceil

from sqlalchemy import orm, and_, or_, inspect
from sqlalchemy.orm.strategy_options import Load

DEFAULT_PER_PAGE = 50
DEFAULT_SIMPLE_SEARCH_KEY = 'q'

class Query(orm.Query):

    # when used as a query property (e.g. MyModel.query), then this is populated with the originating model
    __model__ = None

    @property
    def Model(self):
        '''Return originating model class when Query used a query property'''
        return self.__model__

    @property
    def entities(self):
        '''Return list of entity classes present in query'''
        return [e.mapper.class_ for e in self._entities]

    @property
    def join_entities(self):
        '''Return list of the joined entity classes present in query'''
        return [e.mapper.class_ for e in self._join_entities]

    @property
    def all_entities(self):
        '''Return list of entities + join_entities present in query'''
        return self.entities + self.join_entities

    def _join_eager(self, keys, outerjoin, **kargs):
        '''Helper method for applying join()/outerjoin() with contains_eager()'''
        alias = kargs.pop('alias', None)

        key = keys[0]
        path_keys = keys[1:]

        join_args = ([alias, key] if alias else [key]) + list(path_keys)

        opt = orm.contains_eager(key, alias=alias)

        for k in path_keys:
            opt = opt.contains_eager(k)

        join = self.outerjoin if outerjoin else self.join

        return join(*join_args).options(opt)

    def join_eager(self, *keys, **kargs):
        '''Apply `self.join()` + `self.options(contains_eager())`.'''
        return self._join_eager(keys, False, **kargs)

    def outerjoin_eager(self, *keys, **kargs):
        '''Apply `self.outerjoin()` + `self.options(contains_eager())`.'''
        return self._join_eager(keys, True, **kargs)

    def _join_load(self, keys, load_type, **kargs):
        opt = getattr(orm, load_type)(keys[0], **kargs)

        for k in keys[1:]:
            opt = getattr(opt, load_type)(k)

        return self.options(opt)

    def joinedload(self, *keys, **kargs):
        '''Apply joinedload() to `keys`'''
        return self._join_load(keys, 'joinedload', **kargs)

    def immediateload(self, *keys, **kargs):
        '''Apply immediateload() to `keys`'''
        return self._join_load(keys, 'immediateload', **kargs)

    def lazyload(self, *keys, **kargs):
        '''Apply lazyload() to `keys`'''
        return self._join_load(keys, 'lazyload', **kargs)

    def noload(self, *keys, **kargs):
        '''Apply noload() to `keys`'''
        return self._join_load(keys, 'noload', **kargs)

    def subqueryload(self, *keys, **kargs):
        '''Apply subqueryload() to `keys`'''
        return self._join_load(keys, 'subqueryload', **kargs)

    def load_only(self, *columns):
        '''Apply `load_only()` to query'''
        obj, columns = get_load_options(*columns)
        return self.options(obj.load_only(*columns))

    def defer(self, *columns):
        '''Apply `defer()` to query'''
        obj, columns = get_load_options(*columns)
        return self.options(reduce(lambda result, i: result.defer(i), columns, obj))

    def undefer(self, *columns):
        '''Apply `undefer()` to query'''
        obj, columns = get_load_options(*columns)
        return self.options(reduce(lambda result, i: result.undefer(i), columns, obj))

    def undefer_group(self, *names):
        '''Apply `undefer_group()` to query'''
        obj, names = get_load_options(*names)
        return self.options(obj.undefer_group(names[0]))

    def map(self, function, *iterable_args):
        '''Call native `map` on `self.all()`'''
        return map(function, self.all(), *iterable_args)

    def reduce(self, function, initial=None):
        '''Call native `reduce` on `self.all()`'''
        return reduce(function, self.all(), initial)

    def reduce_right(self, function, initial=None):
        '''Call native `reduce` on reversed `self.all()`'''
        return reduce(function, reversed(self.all()), initial)

    def pluck(self, column):
        '''Pluck `column` attribute values from `self.all()` results and return as list'''
        return [getattr(r, column, None) for r in self.all()]

    def page(self, page=1, per_page=DEFAULT_PER_PAGE):
        '''Return query with limit and offset applied for page'''
        return self.limit(per_page).offset((page - 1) * per_page)

    def paginate(self, page=1, per_page=DEFAULT_PER_PAGE, error_out=True):
        if error_out and page < 1:
            raise IndexError

        items = self.page(page, per_page).all()

        if not items and page != 1 and error_out:
           raise IndexError

        # No need to count if we're on the first page and there are fewer items than we expected.
        if page == 1 and len(items) < per_page:
            total = len(items)
        else:
            total = self.order_by(None).count()

        return Pagination(self, page, per_page, total, items)

    def advanced_search(self, search_dict):
        filters = [m for m in [model.advanced_search(search_dict) for model in self.all_entities] if m is not None]
        return self.filter(and_(*filters)) if filters else self

    def simple_search(self, search_string):
        filters = [m for m in [model.simple_search(search_string) for model in self.all_entities] if m is not None]
        return self.filter(or_(*filters)) if filters else self

    def search(self, search_dict, simple_search_key=DEFAULT_SIMPLE_SEARCH_KEY):
        return self.advanced_search(search_dict).simple_search(search_dict.get(simple_search_key, ''))

class Pagination(object):
    '''
    Internal helper class returned by :meth:`BaseQuery.paginate`.  You
    can also construct it from any other SQLAlchemy query object if you are
    working with other libraries.  Additionally it is possible to pass `None`
    as query object in which case the :meth:`prev` and :meth:`next` will
    no longer work.
    '''

    def __init__(self, query, page, per_page, total, items):
        # the unlimited query object that was used to create this pagination object
        self.query = query
        # the current page number (1 indexed)
        self.page = page
        # the number of items to be displayed on a page.
        self.per_page = per_page
        # the total number of items matching the query
        self.total = total
        # the items for the current page
        self.items = items

        # the total number of pages
        self.pages = 0 if self.per_page == 0 else int(ceil(self.total / float(self.per_page)))
        # number of the previous page
        self.prev_num = self.page - 1
        # True if a previous page exists
        self.has_prev = self.page > 1
        # number of the next page
        self.next_num = self.page + 1
        # True if a next page exists
        self.has_next = self.page < self.pages

    def prev(self, error_out=False):
        '''Returns a :class:`Pagination` object for the previous page.'''
        assert self.query is not None, 'a query object is required for this method to work'
        return self.query.paginate(self.page - 1, self.per_page, error_out)

    def next(self, error_out=False):
        '''Returns a :class:`Pagination` object for the next page.'''
        assert self.query is not None, 'a query object is required for this method to work'
        return self.query.paginate(self.page + 1, self.per_page, error_out)

def get_load_options(*columns):
    '''
    Attempt to extract a sqlalchemy object from `columns[0]` and return remaining columns to apply to a query load method
    '''
    model_inspect = inspect(columns[0], raiseerr=False)

    # return an obj which has loading API
    if model_inspect and model_inspect.is_mapper:
        obj = Load(columns[0])
        columns = columns[1:]
    elif isinstance(columns[0], Load):
        obj = columns[0]
        columns = columns[1:]
    else:
        obj = orm

    return (obj, columns)

